@@ -1,4 +1,23 @@
 /* vim: set sw=4 ts=4 sts=4 et: */
+
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+
+#include <stdio.h>
+#include <string.h>
+#include <ctype.h>
+#include <sys/types.h>
+#include <unistd.h>
+
+#ifdef _WIN32
+# include <winsock2.h>
+#endif
+
+#include <Ecore.h>
+#include <Ecore_Str.h>
+#include <Ecore_File.h>
+
 #include "Efreet.h"
 #include "efreet_private.h"
 
@@ -13,21 +32,27 @@ static const char *desktop_environment = NULL;
  * A cache of all loaded desktops, hashed by file name.
  * Values are Efreet_Desktop structures
  */
-static Ecore_Hash *efreet_desktop_cache = NULL;
+static Eina_Hash *efreet_desktop_cache = NULL;
 
 /**
  * A list of the desktop types available
  */
-static Ecore_List *efreet_desktop_types = NULL;
+static Eina_List *efreet_desktop_types = NULL;
 
 /**
  * A unique id for each tmp file created while building a command
  */
 static int efreet_desktop_command_file_id = 0;
 
-static int init = 0;
 static int cache_flush = 0;
 
+#ifdef EFREET_MODULE_LOG_DOM 
+#undef EFREET_MODULE_LOG_DOM
+#endif
+
+#define EFREET_MODULE_LOG_DOM _efreet_desktop_log_dom
+static int _efreet_desktop_log_dom = -1;
+
 EAPI int EFREET_DESKTOP_TYPE_APPLICATION = 0;
 EAPI int EFREET_DESKTOP_TYPE_LINK = 0;
 EAPI int EFREET_DESKTOP_TYPE_DIRECTORY = 0;
@@ -61,16 +86,20 @@ static int efreet_desktop_generic_fields_parse(Efreet_Desktop *desktop,
                                                 Efreet_Ini *ini);
 static void efreet_desktop_generic_fields_save(Efreet_Desktop *desktop,
                                                 Efreet_Ini *ini);
-static void efreet_desktop_x_fields_parse(Ecore_Hash_Node *node,
-                                            Efreet_Desktop *desktop);
-static void efreet_desktop_x_fields_save(Ecore_Hash_Node *node,
-                                            Efreet_Ini *ini);
+static Eina_Bool efreet_desktop_x_fields_parse(const Eina_Hash *hash,
+					       const void *key,
+					       void *data,
+					       void *fdata);
+static Eina_Bool efreet_desktop_x_fields_save(const Eina_Hash *hash,
+					      const void *key,
+					      void *value,
+					      void *fdata);
 static int efreet_desktop_environment_check(Efreet_Ini *ini);
 static char *efreet_string_append(char *dest, int *size,
                                     int *len, const char *src);
 static char *efreet_string_append_char(char *dest, int *size,
                                         int *len, char c);
-static Ecore_List *efreet_desktop_command_build(Efreet_Desktop_Command *command);
+static Eina_List *efreet_desktop_command_build(Efreet_Desktop_Command *command);
 static void efreet_desktop_command_free(Efreet_Desktop_Command *command);
 static char *efreet_desktop_command_append_quoted(char *dest, int *size,
                                                     int *len, char *src);
@@ -97,12 +126,12 @@ static int efreet_desktop_cb_download_progress(void *data, const char *file,
                                            long int ultotal, long int ulnow);
 
 
-static void efreet_desktop_exec_cb(void *data, Efreet_Desktop *desktop,
+static void *efreet_desktop_exec_cb(void *data, Efreet_Desktop *desktop,
                                             char *exec, int remaining);
 
 static void efreet_desktop_type_info_free(Efreet_Desktop_Type_Info *info);
 static int efreet_desktop_command_flags_get(Efreet_Desktop *desktop);
-static void efreet_desktop_command_execs_process(Efreet_Desktop_Command *command, Ecore_List *execs);
+static void *efreet_desktop_command_execs_process(Efreet_Desktop_Command *command, Eina_List *execs);
 
 /**
  * @internal
@@ -112,16 +141,20 @@ static void efreet_desktop_command_execs_process(Efreet_Desktop_Command *command
 int
 efreet_desktop_init(void)
 {
-    if (init++) return init;
-    if (!ecore_string_init()) return --init;
-    if (!ecore_file_init()) return --init;
+    _efreet_desktop_log_dom = eina_log_domain_register("Efreet_desktop", EFREET_DEFAULT_LOG_COLOR);
+    if (_efreet_desktop_log_dom < 0)
+    {
+	ERROR("Efreet: Could not create a log domain for Efreet_desktop");
+	return 0;
+    }
+    if (!ecore_file_init())
+    {
+        eina_log_domain_unregister(_efreet_desktop_log_dom);
+        return 0;
+    }
 
-    efreet_desktop_cache = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-    ecore_hash_free_key_cb_set(efreet_desktop_cache, ECORE_FREE_CB(free));
-
-    efreet_desktop_types = ecore_list_new();
-    ecore_list_free_cb_set(efreet_desktop_types,
-                            ECORE_FREE_CB(efreet_desktop_type_info_free));
+    efreet_desktop_cache = eina_hash_string_superfast_new(NULL);
+    efreet_desktop_types = NULL;
 
     EFREET_DESKTOP_TYPE_APPLICATION = efreet_desktop_type_add("Application",
                                         efreet_desktop_application_fields_parse,
@@ -133,7 +166,7 @@ efreet_desktop_init(void)
     EFREET_DESKTOP_TYPE_DIRECTORY = efreet_desktop_type_add("Directory", NULL,
                                                                 NULL, NULL);
 
-    return init;
+    return 1;
 }
 
 /**
@@ -141,18 +174,22 @@ efreet_desktop_init(void)
  * @returns the number of initializations left for this system
  * @brief Attempts to shut down the subsystem if nothing else is using it
  */
-int
+void
 efreet_desktop_shutdown(void)
 {
-    if (--init) return init;
-    ecore_file_shutdown();
-    ecore_string_shutdown();
+    Efreet_Desktop_Type_Info *info;
 
     IF_RELEASE(desktop_environment);
     IF_FREE_HASH(efreet_desktop_cache);
-    IF_FREE_LIST(efreet_desktop_types);
-
-    return init;
+    while (efreet_desktop_types)
+    {
+        info = eina_list_data_get(efreet_desktop_types);
+        efreet_desktop_type_info_free(info);
+        efreet_desktop_types = eina_list_remove_list(efreet_desktop_types,
+                                                     efreet_desktop_types);
+    }
+    ecore_file_shutdown();
+    eina_log_domain_unregister(_efreet_desktop_log_dom);
 }
 
 /**
@@ -188,7 +225,7 @@ efreet_desktop_get(const char *file)
 
     if (efreet_desktop_cache)
     {
-        desktop = ecore_hash_get(efreet_desktop_cache, file);
+        desktop = eina_hash_find(efreet_desktop_cache, file);
         if (desktop)
         {
             if (efreet_desktop_cache_check(desktop))
@@ -206,14 +243,14 @@ efreet_desktop_get(const char *file)
             }
 
             desktop->cached = 0;
-            ecore_hash_remove(efreet_desktop_cache, file);
+            eina_hash_del(efreet_desktop_cache, file, NULL);
         }
     }
 
     desktop = efreet_desktop_new(file);
     if (!desktop) return NULL;
 
-    ecore_hash_set(efreet_desktop_cache, strdup(file), desktop);
+    eina_hash_add(efreet_desktop_cache, file, desktop);
     desktop->cached = 1;
     return desktop;
 }
@@ -306,7 +343,7 @@ efreet_desktop_read(Efreet_Desktop *desktop)
     if (!ok) ok = efreet_ini_section_set(ini, "KDE Desktop Entry");
     if (!ok)
     {
-        printf("efreet_desktop_new error: no Desktop Entry section\n");
+        ERR("efreet_desktop_new error: no Desktop Entry section");
         error = 1;
     }
 
@@ -330,8 +367,7 @@ efreet_desktop_read(Efreet_Desktop *desktop)
     if (!error && !efreet_desktop_environment_check(ini)) error = 1;
     if (!error && !efreet_desktop_generic_fields_parse(desktop, ini)) error = 1;
     if (!error)
-        ecore_hash_for_each_node(ini->section,
-                ECORE_FOR_EACH(efreet_desktop_x_fields_parse), desktop);
+       eina_hash_foreach(ini->section, efreet_desktop_x_fields_parse, desktop);
 
     efreet_ini_free(ini);
 
@@ -351,6 +387,8 @@ efreet_desktop_read(Efreet_Desktop *desktop)
 static void
 efreet_desktop_clear(Efreet_Desktop *desktop)
 {
+    char *data;
+
     IF_FREE(desktop->name);
     IF_FREE(desktop->generic_name);
     IF_FREE(desktop->comment);
@@ -362,17 +400,27 @@ efreet_desktop_clear(Efreet_Desktop *desktop)
     IF_FREE(desktop->path);
     IF_FREE(desktop->startup_wm_class);
 
-    IF_FREE_LIST(desktop->only_show_in);
-    IF_FREE_LIST(desktop->not_show_in);
-    IF_FREE_LIST(desktop->categories);
-    IF_FREE_LIST(desktop->mime_types);
+    IF_FREE_LIST(desktop->only_show_in, free);
+    IF_FREE_LIST(desktop->not_show_in, free);
+    while (desktop->categories)
+    {
+        data = eina_list_data_get(desktop->categories);
+        eina_stringshare_del(data);
+        desktop->categories = eina_list_remove_list(desktop->categories, desktop->categories);
+    }
+    while (desktop->mime_types)
+    {
+        data = eina_list_data_get(desktop->mime_types);
+        eina_stringshare_del(data);
+        desktop->mime_types = eina_list_remove_list(desktop->mime_types, desktop->mime_types);
+    }
 
     IF_FREE_HASH(desktop->x);
 
     if (desktop->type_data)
     {
         Efreet_Desktop_Type_Info *info;
-        info = ecore_list_index_goto(efreet_desktop_types, desktop->type);
+        info = eina_list_nth(efreet_desktop_types, desktop->type);
         if (info->free_func)
             info->free_func(desktop->type_data);
     }
@@ -391,11 +439,11 @@ efreet_desktop_save(Efreet_Desktop *desktop)
     Efreet_Ini *ini;
     int ok = 1;
 
-    ini = efreet_ini_new(desktop->orig_path);
+    ini = efreet_ini_new(NULL);
     efreet_ini_section_add(ini, "Desktop Entry");
     efreet_ini_section_set(ini, "Desktop Entry");
 
-    info = ecore_list_index_goto(efreet_desktop_types, desktop->type);
+    info = eina_list_nth(efreet_desktop_types, desktop->type);
     if (info)
     {
         efreet_ini_string_set(ini, "Type", info->type);
@@ -428,11 +476,12 @@ efreet_desktop_save(Efreet_Desktop *desktop)
         if (!efreet_ini_save(ini, desktop->orig_path)) ok = 0;
         else
         {
-            if (desktop != ecore_hash_get(efreet_desktop_cache, desktop->orig_path))
+            if (desktop != eina_hash_find(efreet_desktop_cache, desktop->orig_path))
             {
                 desktop->cached = 1;
-                ecore_hash_set(efreet_desktop_cache,
-                                strdup(desktop->orig_path), desktop);
+                eina_hash_del(efreet_desktop_cache, desktop->orig_path, NULL);
+                eina_hash_add(efreet_desktop_cache, desktop->orig_path,
+                              desktop);
             }
         }
     }
@@ -449,10 +498,10 @@ efreet_desktop_save(Efreet_Desktop *desktop)
 EAPI int
 efreet_desktop_save_as(Efreet_Desktop *desktop, const char *file)
 {
-    if (desktop == ecore_hash_get(efreet_desktop_cache, desktop->orig_path))
+    if (desktop == eina_hash_find(efreet_desktop_cache, desktop->orig_path))
     {
         desktop->cached = 0;
-        ecore_hash_remove(efreet_desktop_cache, desktop->orig_path);
+        eina_hash_del(efreet_desktop_cache, desktop->orig_path, NULL);
     }
     FREE(desktop->orig_path);
     desktop->orig_path = strdup(file);
@@ -468,13 +517,15 @@ efreet_desktop_save_as(Efreet_Desktop *desktop, const char *file)
 EAPI void
 efreet_desktop_free(Efreet_Desktop *desktop)
 {
+    char *str;
+
     if (!desktop) return;
 
     desktop->ref--;
     if (desktop->ref > 0) return;
 
     if (desktop->cached && efreet_desktop_cache)
-        ecore_hash_remove(efreet_desktop_cache, desktop->orig_path);
+      eina_hash_del(efreet_desktop_cache, desktop->orig_path, NULL);
 
     IF_FREE(desktop->orig_path);
 
@@ -489,17 +540,20 @@ efreet_desktop_free(Efreet_Desktop *desktop)
     IF_FREE(desktop->path);
     IF_FREE(desktop->startup_wm_class);
 
-    IF_FREE_LIST(desktop->only_show_in);
-    IF_FREE_LIST(desktop->not_show_in);
-    IF_FREE_LIST(desktop->categories);
-    IF_FREE_LIST(desktop->mime_types);
+    IF_FREE_LIST(desktop->only_show_in, free);
+    IF_FREE_LIST(desktop->not_show_in, free);
+
+    EINA_LIST_FREE(desktop->categories, str)
+        eina_stringshare_del(str);
+    EINA_LIST_FREE(desktop->mime_types, str)
+        eina_stringshare_del(str);
 
     IF_FREE_HASH(desktop->x);
 
     if (desktop->type_data)
     {
         Efreet_Desktop_Type_Info *info;
-        info = ecore_list_index_goto(efreet_desktop_types, desktop->type);
+        info = eina_list_nth(efreet_desktop_types, desktop->type);
         if (info->free_func)
             info->free_func(desktop->type_data);
     }
@@ -510,22 +564,25 @@ efreet_desktop_free(Efreet_Desktop *desktop)
 /**
  * @param desktop: The desktop file to work with
  * @param files: The files to be substituted into the exec line
- * @aparam data: The data pointer to pass
+ * @param data: The data pointer to pass
  * @return Returns the Ecore_Exce for @a desktop
  * @brief Parses the @a desktop exec line and returns an Ecore_Exe.
  */
 EAPI void
-efreet_desktop_exec(Efreet_Desktop *desktop, Ecore_List *files, void *data)
+efreet_desktop_exec(Efreet_Desktop *desktop, Eina_List *files, void *data)
 {
     efreet_desktop_command_get(desktop, files, efreet_desktop_exec_cb, data);
 }
 
-static void
+static void *
 efreet_desktop_exec_cb(void *data, Efreet_Desktop *desktop __UNUSED__,
                                 char *exec, int remaining __UNUSED__)
 {
+#ifndef _WIN32
     ecore_exe_run(exec, data);
     free(exec);
+#endif
+    return NULL;
 }
 
 /**
@@ -535,8 +592,8 @@ efreet_desktop_exec_cb(void *data, Efreet_Desktop *desktop __UNUSED__,
 EAPI void
 efreet_desktop_environment_set(const char *environment)
 {
-    if (desktop_environment) ecore_string_release(desktop_environment);
-    if (environment) desktop_environment = ecore_string_instance(environment);
+    if (desktop_environment) eina_stringshare_del(desktop_environment);
+    if (environment) desktop_environment = eina_stringshare_add(environment);
     else desktop_environment = NULL;
 }
 
@@ -560,7 +617,7 @@ EAPI unsigned int
 efreet_desktop_category_count_get(Efreet_Desktop *desktop)
 {
     if (!desktop || !desktop->categories) return 0;
-    return ecore_list_count(desktop->categories);
+    return eina_list_count(desktop->categories);
 }
 
 /**
@@ -573,18 +630,11 @@ efreet_desktop_category_add(Efreet_Desktop *desktop, const char *category)
 {
     if (!desktop) return;
 
-    if (!desktop->categories)
-    {
-        desktop->categories = ecore_list_new();
-        ecore_list_free_cb_set(desktop->categories,
-                                ECORE_FREE_CB(ecore_string_release));
-    }
-    else
-        if (ecore_list_find(desktop->categories,
-                                ECORE_COMPARE_CB(strcmp), category)) return;
+    if (eina_list_search_unsorted(desktop->categories,
+                                  EINA_COMPARE_CB(strcmp), category)) return;
 
-    ecore_list_append(desktop->categories,
-                        (void *)ecore_string_instance(category));
+    desktop->categories = eina_list_append(desktop->categories,
+                        (void *)eina_stringshare_add(category));
 }
 
 /**
@@ -596,17 +646,20 @@ efreet_desktop_category_add(Efreet_Desktop *desktop, const char *category)
 EAPI int
 efreet_desktop_category_del(Efreet_Desktop *desktop, const char *category)
 {
-    int found = 0;
+    char *found = NULL;
+
     if (!desktop || !desktop->categories) return 0;
 
-    if (ecore_list_find(desktop->categories,
-                            ECORE_COMPARE_CB(strcmp), category))
+    if ((found = eina_list_search_unsorted(desktop->categories,
+                                           EINA_COMPARE_CB(strcmp), category)))
     {
-        found = 1;
-        ecore_list_remove(desktop->categories);
+        eina_stringshare_del(found);
+        desktop->categories = eina_list_remove(desktop->categories, found);
+
+        return 1;
     }
 
-    return found;
+    return 0;
 }
 
 /**
@@ -628,7 +681,7 @@ efreet_desktop_type_add(const char *type, Efreet_Desktop_Type_Parse_Cb parse_fun
     info = NEW(Efreet_Desktop_Type_Info, 1);
     if (!info) return 0;
 
-    id = ecore_list_count(efreet_desktop_types);
+    id = eina_list_count(efreet_desktop_types);
 
     info->id = id;
     info->type = strdup(type);
@@ -636,7 +689,7 @@ efreet_desktop_type_add(const char *type, Efreet_Desktop_Type_Parse_Cb parse_fun
     info->save_func = save_func;
     info->free_func = free_func;
 
-    ecore_list_append(efreet_desktop_types, info);
+    efreet_desktop_types = eina_list_append(efreet_desktop_types, info);
 
     return id;
 }
@@ -653,7 +706,7 @@ EAPI int
 efreet_desktop_type_alias(int from_type, const char *alias)
 {
     Efreet_Desktop_Type_Info *info;
-    info = ecore_list_index_goto(efreet_desktop_types, from_type);
+    info = eina_list_nth(efreet_desktop_types, from_type);
     if (!info) return -1;
 
     return efreet_desktop_type_add(alias, info->parse_func, info->save_func, info->free_func);
@@ -692,11 +745,11 @@ static Efreet_Desktop_Type_Info *
 efreet_desktop_type_parse(const char *type_str)
 {
     Efreet_Desktop_Type_Info *info;
+    Eina_List *l;
 
     if (!type_str) return NULL;
 
-    ecore_list_first_goto(efreet_desktop_types);
-    while ((info = ecore_list_next(efreet_desktop_types)))
+    EINA_LIST_FOREACH(efreet_desktop_types, l, info)
     {
         if (!strcmp(info->type, type_str))
             return info;
@@ -707,23 +760,18 @@ efreet_desktop_type_parse(const char *type_str)
 
 /**
  * @param string: the raw string list
- * @return an Ecore_List of ecore string's
+ * @return an Eina_List of ecore string's
  * @brief Parse ';' separate list of strings according to the desktop spec
  */
-EAPI Ecore_List *
+EAPI Eina_List *
 efreet_desktop_string_list_parse(const char *string)
 {
-    Ecore_List *list;
+    Eina_List *list = NULL;
     char *tmp;
     char *s, *p;
 
     if (!string) return NULL;
 
-    list = ecore_list_new();
-    if (!list) return NULL;
-
-    ecore_list_free_cb_set(list, ECORE_FREE_CB(ecore_string_release));
-
     tmp = strdup(string);
     s = tmp;
 
@@ -731,17 +779,17 @@ efreet_desktop_string_list_parse(const char *string)
     {
         if (p > tmp && *(p-1) == '\\') continue;
         *p = '\0';
-        ecore_list_append(list, (void *)ecore_string_instance(s));
+        list = eina_list_append(list, (void *)eina_stringshare_add(s));
         s = p + 1;
     }
     /* If this is true, the .desktop file does not follow the standard */
     if (*s)
     {
 #ifdef STRICT_SPEC
-        printf("[Efreet]: Found a string list without ';' "
-                "at the end: %s\n", string);
+        WRN("[Efreet]: Found a string list without ';' "
+                "at the end: %s", string);
 #endif
-        ecore_list_append(list, (void *)ecore_string_instance(s));
+        list = eina_list_append(list, (void *)eina_stringshare_add(s));
     }
 
     free(tmp);
@@ -750,25 +798,25 @@ efreet_desktop_string_list_parse(const char *string)
 }
 
 /**
- * @param list: Ecore_List with strings
+ * @param list: Eina_List with strings
  * @return a raw string list
  * @brief Create a ';' separate list of strings according to the desktop spec
  */
 EAPI char *
-efreet_desktop_string_list_join(Ecore_List *list)
+efreet_desktop_string_list_join(Eina_List *list)
 {
+    Eina_List *l;
     const char *tmp;
     char *string;
     size_t size, pos, len;
 
-    if (ecore_list_empty_is(list)) return strdup("");
+    if (!list) return strdup("");
 
     size = 1024;
     string = malloc(size);
     pos = 0;
 
-    ecore_list_first_goto(list);
-    while ((tmp = ecore_list_next(list)))
+    EINA_LIST_FOREACH(list, l, tmp)
     {
         len = strlen(tmp);
         /* +1 for ';' */
@@ -921,7 +969,7 @@ efreet_desktop_generic_fields_parse(Efreet_Desktop *desktop, Efreet_Ini *ini)
     if (val) desktop->name = strdup(val);
     else
     {
-        printf("efreet_desktop_generic_fields_parse error: no Name\n");
+        ERR("efreet_desktop_generic_fields_parse error: no Name");
         return 0;
     }
 
@@ -984,33 +1032,31 @@ efreet_desktop_generic_fields_save(Efreet_Desktop *desktop, Efreet_Ini *ini)
     efreet_ini_boolean_set(ini, "NoDisplay", desktop->no_display);
     efreet_ini_boolean_set(ini, "Hidden", desktop->hidden);
 
-    if (desktop->x) ecore_hash_for_each_node(desktop->x,
-                                             ECORE_FOR_EACH(efreet_desktop_x_fields_save),
-                                             ini);
+    if (desktop->x) eina_hash_foreach(desktop->x, efreet_desktop_x_fields_save,
+				      ini);
 }
 
 /**
  * @internal
  * @param node: The node to work with
  * @param desktop: The desktop file to work with
- * @return Returns no value
+ * @return Returns always true, to be used in eina_hash_foreach()
  * @brief Parses out an X- key from @a node and stores in @a desktop
  */
-static void
-efreet_desktop_x_fields_parse(Ecore_Hash_Node *node, Efreet_Desktop *desktop)
+static Eina_Bool
+efreet_desktop_x_fields_parse(const Eina_Hash *hash __UNUSED__, const void *key, void *value, void *fdata)
 {
-    if (strncmp(node->key, "X-", 2)) return;
+    Efreet_Desktop * desktop = fdata;
+
+    if (!desktop) return EINA_TRUE;
+    if (strncmp(key, "X-", 2)) return EINA_TRUE;
 
     if (!desktop->x)
-    {
-        desktop->x = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-        ecore_hash_free_key_cb_set(desktop->x,
-                            ECORE_FREE_CB(ecore_string_release));
-        ecore_hash_free_value_cb_set(desktop->x,
-                            ECORE_FREE_CB(ecore_string_release));
-    }
-    ecore_hash_set(desktop->x, (void *)ecore_string_instance(node->key),
-            (void *)ecore_string_instance(node->value));
+      desktop->x = eina_hash_string_superfast_new(EINA_FREE_CB(eina_stringshare_del));
+    eina_hash_del(desktop->x, key, NULL);
+    eina_hash_add(desktop->x, key, (void *)eina_stringshare_add(value));
+
+    return EINA_TRUE;
 }
 
 /**
@@ -1020,10 +1066,13 @@ efreet_desktop_x_fields_parse(Ecore_Hash_Node *node, Efreet_Desktop *desktop)
  * @return Returns no value
  * @brief Stores an X- key from @a node and stores in @a ini
  */
-static void
-efreet_desktop_x_fields_save(Ecore_Hash_Node *node, Efreet_Ini *ini)
+static Eina_Bool
+efreet_desktop_x_fields_save(const Eina_Hash *hash __UNUSED__, const void *key, void *value, void *fdata)
 {
-    efreet_ini_string_set(ini, node->key, node->value);
+    Efreet_Ini *ini = fdata;
+    efreet_ini_string_set(ini, key, value);
+
+    return EINA_TRUE;
 }
 
 
@@ -1037,68 +1086,49 @@ efreet_desktop_x_fields_save(Ecore_Hash_Node *node, Efreet_Ini *ini)
 static int
 efreet_desktop_environment_check(Efreet_Ini *ini)
 {
-    Ecore_List *list;
-    const char *val;
+    Eina_List *list;
+    int found = 0;
+    char *val;
+
+    if (!desktop_environment)
+      return 1;
 
     list = efreet_desktop_string_list_parse(efreet_ini_string_get(ini, "OnlyShowIn"));
     if (list)
     {
-        int found = 0;
-
-        if (desktop_environment)
-        {
-            ecore_list_first_goto(list);
-            while ((val = ecore_list_next(list)))
+       EINA_LIST_FREE(list, val)
             {
                 if (!strcmp(val, desktop_environment))
-                {
                     found = 1;
-                    break;
-                }
-            }
+	    eina_stringshare_del(val);
         }
 
-        ecore_list_destroy(list);
         return found;
     }
 
-    if (desktop_environment)
-    {
-        int found = 0;
-
         list = efreet_desktop_string_list_parse(efreet_ini_string_get(ini, "NotShowIn"));
-        if (list)
-        {
-            ecore_list_first_goto(list);
-            while ((val = ecore_list_next(list)))
+    EINA_LIST_FREE(list, val)
             {
                 if (!strcmp(val, desktop_environment))
-                {
                     found = 1;
-                    break;
-                }
-            }
-            ecore_list_destroy(list);
+	 eina_stringshare_del(val);
         }
 
         return !found;
-    }
-
-    return 1;
 }
 
 
 /**
  * @param desktop: the desktop entry
- * @param files: an ecore list of file names to execute, as either absolute paths,
+ * @param files: an eina list of file names to execute, as either absolute paths,
  * relative paths, or uris
  * @param func: a callback to call for each prepared command line
  * @param data: user data passed to the callback
- * @return Returns 1 on success or 0 on failure
+ * @return Returns the return value of @p func on success or NULL on failure
  * @brief Get a command to use to execute a desktop entry.
  */
-EAPI int
-efreet_desktop_command_get(Efreet_Desktop *desktop, Ecore_List *files,
+EAPI void *
+efreet_desktop_command_get(Efreet_Desktop *desktop, Eina_List *files,
                             Efreet_Desktop_Command_Cb func, void *data)
 {
     return efreet_desktop_command_progress_get(desktop, files, func, NULL, data);
@@ -1106,36 +1136,31 @@ efreet_desktop_command_get(Efreet_Desktop *desktop, Ecore_List *files,
 
 /**
  * @param desktop: the desktop entry
- * @param files an ecore list of local files, as absolute paths, local paths, or file:// uris (or NULL to get exec string with no files appended)
- * @return Returns an ecore list of exec strings
+ * @param files an eina list of local files, as absolute paths, local paths, or file:// uris (or NULL to get exec string with no files appended)
+ * @return Returns an eina list of exec strings
  * @brief Get the command to use to execute a desktop entry
  *
  * The returned list and each of its elements must be freed.
  */
-EAPI Ecore_List *
-efreet_desktop_command_local_get(Efreet_Desktop *desktop, Ecore_List *files)
+EAPI Eina_List *
+efreet_desktop_command_local_get(Efreet_Desktop *desktop, Eina_List *files)
 {
     Efreet_Desktop_Command *command;
     char *file;
-    Ecore_List *execs;
+    Eina_List *execs, *l;
 
     if (!desktop || !desktop->exec) return NULL;
 
     command = NEW(Efreet_Desktop_Command, 1);
     if (!command) return 0;
 
-    command->files = ecore_list_new();
     command->desktop = desktop;
 
-    ecore_list_free_cb_set(command->files,
-                            ECORE_FREE_CB(efreet_desktop_command_file_free));
-
     command->flags = efreet_desktop_command_flags_get(desktop);
     /* get the required info for each file passed in */
     if (files)
     {
-        ecore_list_first_goto(files);
-        while ((file = ecore_list_next(files)))
+        EINA_LIST_FOREACH(files, l, file)
         {
             Efreet_Desktop_Command_File *dcf;
 
@@ -1146,7 +1171,7 @@ efreet_desktop_command_local_get(Efreet_Desktop *desktop, Ecore_List *files)
                 efreet_desktop_command_file_free(dcf);
                 continue;
             }
-            ecore_list_append(command->files, dcf);
+            command->files = eina_list_append(command->files, dcf);
         }
     }
 
@@ -1159,7 +1184,7 @@ efreet_desktop_command_local_get(Efreet_Desktop *desktop, Ecore_List *files)
 
 /**
  * @param desktop: the desktop entry
- * @param files: an ecore list of file names to execute, as either absolute paths,
+ * @param files: an eina list of file names to execute, as either absolute paths,
  * relative paths, or uris
  * @param cb_command: a callback to call for each prepared command line
  * @param cb_progress: a callback to get progress for the downloads
@@ -1168,55 +1193,53 @@ efreet_desktop_command_local_get(Efreet_Desktop *desktop, Ecore_List *files)
  * @brief Get a command to use to execute a desktop entry, and receive progress
  * updates for downloading of remote URI's passed in.
  */
-EAPI int
-efreet_desktop_command_progress_get(Efreet_Desktop *desktop, Ecore_List *files,
+EAPI void *
+efreet_desktop_command_progress_get(Efreet_Desktop *desktop, Eina_List *files,
                                     Efreet_Desktop_Command_Cb cb_command,
                                     Efreet_Desktop_Progress_Cb cb_progress,
                                     void *data)
 {
     Efreet_Desktop_Command *command;
+    Eina_List *l;
     char *file;
+    void *ret = NULL;
 
-    if (!desktop || !cb_command || !desktop->exec) return 0;
+    if (!desktop || !cb_command || !desktop->exec) return NULL;
 
     command = NEW(Efreet_Desktop_Command, 1);
-    if (!command) return 0;
+    if (!command) return NULL;
 
     command->cb_command = cb_command;
     command->cb_progress = cb_progress;
     command->data = data;
-    command->files = ecore_list_new();
     command->desktop = desktop;
 
-    ecore_list_free_cb_set(command->files,
-                            ECORE_FREE_CB(efreet_desktop_command_file_free));
-
     command->flags = efreet_desktop_command_flags_get(desktop);
     /* get the required info for each file passed in */
     if (files)
     {
-        ecore_list_first_goto(files);
-        while ((file = ecore_list_next(files)))
+        EINA_LIST_FOREACH(files, l, file)
         {
             Efreet_Desktop_Command_File *dcf;
 
             dcf = efreet_desktop_command_file_process(command, file);
             if (!dcf) continue;
-            ecore_list_append(command->files, dcf);
+            command->files = eina_list_append(command->files, dcf);
             command->num_pending += dcf->pending;
         }
     }
 
     if (command->num_pending == 0)
     {
-        Ecore_List *execs;
+        Eina_List *execs;
+
         execs = efreet_desktop_command_build(command);
-        efreet_desktop_command_execs_process(command, execs);
-        ecore_list_destroy(execs);
+        ret = efreet_desktop_command_execs_process(command, execs);
+        eina_list_free(execs);
         efreet_desktop_command_free(command);
     }
 
-    return 1;
+    return ret;
 }
 
 /**
@@ -1283,17 +1306,20 @@ efreet_desktop_command_flags_get(Efreet_Desktop *desktop)
  * @param command
  * @param execs
  */
-static void
-efreet_desktop_command_execs_process(Efreet_Desktop_Command *command, Ecore_List *execs)
+static void *
+efreet_desktop_command_execs_process(Efreet_Desktop_Command *command, Eina_List *execs)
 {
+    Eina_List *l;
     char *exec;
     int num;
-    num = ecore_list_count(execs);
-    ecore_list_first_goto(execs);
-    while ((exec = ecore_list_next(execs)))
+    void *ret = NULL;
+   
+    num = eina_list_count(execs);
+    EINA_LIST_FOREACH(execs, l, exec)
     {
-        command->cb_command(command->data, command->desktop, exec, --num);
+        ret = command->cb_command(command->data, command->desktop, exec, --num);
     }
+    return ret;
 }
 
 
@@ -1305,30 +1331,25 @@ efreet_desktop_command_execs_process(Efreet_Desktop_Command *command, Ecore_List
  * @param command: the command to build
  * @return a list of executable strings
  */
-static Ecore_List *
+static Eina_List *
 efreet_desktop_command_build(Efreet_Desktop_Command *command)
 {
-    Efreet_Desktop_Command_File *file = NULL;
-    int first = 1;
-    Ecore_List *execs;
+    Eina_List *execs = NULL;
+    const Eina_List *l;
     char *exec;
 
-    execs = ecore_list_new();
-
-    ecore_list_first_goto(command->files);
-
     /* if the Exec field appends multiple, that will run the list to the end,
      * causing this loop to only run once. otherwise, this loop will generate a
      * command for each file in the list. if the list is empty, this
      * will run once, removing any file field codes */
-    while ((file = ecore_list_next(command->files)) || first)
-    {
+    l = command->files;
+    do
+      {
         const char *p;
         int len = 0;
         int size = PATH_MAX;
         int file_added = 0;
-
-        first = 0;
+	Efreet_Desktop_Command_File *file = eina_list_data_get(l);
 
         exec = malloc(size);
         p = command->desktop->exec;
@@ -1384,7 +1405,7 @@ efreet_desktop_command_build(Efreet_Desktop_Command *command)
                         break;
                     case 'v':
                     case 'm':
-                        printf("[Efreet]: Deprecated conversion char: '%c' in file '%s'\n",
+                        WRN("[Efreet]: Deprecated conversion char: '%c' in file '%s'",
                                                             *p, command->desktop->orig_path);
                         break;
                     case '%':
@@ -1392,7 +1413,7 @@ efreet_desktop_command_build(Efreet_Desktop_Command *command)
                         break;
                     default:
 #ifdef STRICT_SPEC
-                        printf("[Efreet]: Unknown conversion character: '%c'\n", *p);
+                        WRN("[Efreet_desktop]: Unknown conversion character: '%c'", *p);
 #endif
                         break;
                 }
@@ -1409,7 +1430,7 @@ efreet_desktop_command_build(Efreet_Desktop_Command *command)
 	*/
        if ((file) && (!file_added))
 	 {
-	    printf("[Efreet]: %s\n"
+	    WRN("Efreet_desktop: %s\n"
 		   "  command: %s\n"
 		   "  has no file path/uri spec info for executing this app WITH a\n"
 		   "  file/uri as a parameter. This is unlikely to be the intent.\n"
@@ -1429,12 +1450,13 @@ efreet_desktop_command_build(Efreet_Desktop_Command *command)
 #endif
         exec[len++] = '\0';
 
-        ecore_list_append(execs, exec);
+        execs = eina_list_append(execs, exec);
 
         /* If no file was added, then the Exec field doesn't contain any file
          * fields (fFuUdDnN). We only want to run the app once in this case. */
         if (!file_added) break;
-    }
+      }
+    while ((l = eina_list_next(l)) != NULL);
 
     return execs;
 }
@@ -1442,9 +1464,17 @@ efreet_desktop_command_build(Efreet_Desktop_Command *command)
 static void
 efreet_desktop_command_free(Efreet_Desktop_Command *command)
 {
+    Efreet_Desktop_Command_File *dcf;
+
     if (!command) return;
 
-    IF_FREE_LIST(command->files);
+    while (command->files)
+    {
+        dcf = eina_list_data_get(command->files);
+        efreet_desktop_command_file_free(dcf);
+        command->files = eina_list_remove_list(command->files,
+                                               command->files);
+    }
     FREE(command);
 }
 
@@ -1462,7 +1492,7 @@ efreet_desktop_command_append_quoted(char *dest, int *size, int *len, char *src)
         while (*p)
         {
             if (*p == '\'')
-                dest = efreet_string_append_char(dest, size, len, '\\');
+                dest = efreet_string_append(dest, size, len, "\'\\\'");
 
             dest = efreet_string_append_char(dest, size, len, *p);
             p++;
@@ -1482,12 +1512,12 @@ efreet_desktop_command_append_multiple(char *dest, int *size, int *len,
                                         char type)
 {
     Efreet_Desktop_Command_File *file;
+    Eina_List *l;
     int first = 1;
 
     if (!command->files) return dest;
 
-    ecore_list_first_goto(command->files);
-    while ((file = ecore_list_next(command->files)))
+    EINA_LIST_FOREACH(command->files, l, file)
     {
         if (first)
             first = 0;
@@ -1522,8 +1552,8 @@ efreet_desktop_command_append_single(char *dest, int *size, int *len,
             str = file->file;
             break;
         default:
-            printf("Invalid type passed to efreet_desktop_command_append_single:"
-                                                                " '%c'\n", type);
+            ERR("Invalid type passed to efreet_desktop_command_append_single:"
+                                                                " '%c'", type);
             return dest;
     }
 
@@ -1600,7 +1630,7 @@ efreet_desktop_command_file_process(Efreet_Desktop_Command *command, const char
     const char *uri, *base;
     int nonlocal = 0;
 /*
-    printf("FLAGS: %d, %d, %d, %d\n",
+    DBG("FLAGS: %d, %d, %d, %d\n",
         command->flags & EFREET_DESKTOP_EXEC_FLAG_FULLPATH ? 1 : 0,
         command->flags & EFREET_DESKTOP_EXEC_FLAG_URI ? 1 : 0,
         command->flags & EFREET_DESKTOP_EXEC_FLAG_DIR ? 1 : 0,
@@ -1642,7 +1672,7 @@ efreet_desktop_command_file_process(Efreet_Desktop_Command *command, const char
             f->pending = 1;
 
             ecore_file_download(uri, f->fullpath, efreet_desktop_cb_download_complete,
-                                            efreet_desktop_cb_download_progress, f);
+                                            efreet_desktop_cb_download_progress, f, NULL);
         }
 
         if (command->flags & EFREET_DESKTOP_EXEC_FLAG_URI)
@@ -1673,10 +1703,10 @@ efreet_desktop_command_file_process(Efreet_Desktop_Command *command, const char
         free(abs);
     }
 #if 0
-    printf("  fullpath: %s\n", f->fullpath);
-    printf("  uri: %s\n", f->uri);
-    printf("  dir: %s\n", f->dir);
-    printf("  file: %s\n", f->file);
+    INF("  fullpath: %s", f->fullpath);
+    INF("  uri: %s", f->uri);
+    INF("  dir: %s", f->dir);
+    INF("  file: %s", f->file);
 #endif
     return f;
 }
@@ -1751,6 +1781,7 @@ efreet_desktop_cb_download_complete(void *data, const char *file __UNUSED__,
                                                         int status __UNUSED__)
 {
     Efreet_Desktop_Command_File *f;
+
     f = data;
 
     /* XXX check status... error handling, etc */
@@ -1759,10 +1790,12 @@ efreet_desktop_cb_download_complete(void *data, const char *file __UNUSED__,
 
     if (f->command->num_pending <= 0)
     {
-        Ecore_List *execs;
+        Eina_List *execs;
+
         execs = efreet_desktop_command_build(f->command);
+        /* TODO: Need to handle the return value from efreet_desktop_command_execs_process */
         efreet_desktop_command_execs_process(f->command, execs);
-        ecore_list_destroy(execs);
+        eina_list_free(execs);
         efreet_desktop_command_free(f->command);
     }
 }
@@ -1800,8 +1833,12 @@ efreet_desktop_command_path_absolute(const char *path)
     /* relative url */
     if (path[0] != '/')
     {
-        buf = malloc(size);
-        if (!getcwd(buf, size)) return NULL;
+        if (!(buf = malloc(size))) return NULL;
+        if (!getcwd(buf, size))
+        {
+            FREE(buf);
+            return NULL;
+        }
         len = strlen(buf);
 
         if (buf[len-1] != '/') buf = efreet_string_append(buf, &size, &len, "/");
@@ -1810,6 +1847,51 @@ efreet_desktop_command_path_absolute(const char *path)
         return buf;
     }
 
-    /* just dup an alreaady absolute buffer */
+    /* just dup an already absolute buffer */
     return strdup(path);
 }
+
+EAPI Eina_Bool
+efreet_desktop_x_field_set(Efreet_Desktop *desktop, const char *key, const char *data)
+{
+   if (!desktop || strncmp(key, "X-", 2))
+      return EINA_FALSE;
+
+   if (!desktop->x)
+     desktop->x = eina_hash_string_superfast_new(EINA_FREE_CB(eina_stringshare_del));
+   
+   eina_hash_del(desktop->x, key, NULL);
+   eina_hash_add(desktop->x, key, eina_stringshare_add(data));
+   
+   return EINA_TRUE;
+}
+
+EAPI const char *
+efreet_desktop_x_field_get(Efreet_Desktop *desktop, const char *key)
+{
+   const char *ret;
+   
+   if (!desktop || strncmp(key, "X-", 2))
+      return NULL;
+   
+   if (!desktop->x)
+      return NULL;
+   
+   ret = eina_hash_find(desktop->x, key);
+   if (!ret)
+      return NULL;
+   
+   return eina_stringshare_add(ret);
+}
+
+EAPI Eina_Bool
+efreet_desktop_x_field_del(Efreet_Desktop *desktop, const char *key)
+{
+   if (!desktop || strncmp(key, "X-", 2))
+      return EINA_FALSE;
+   
+   if (!desktop->x)
+      return EINA_FALSE;
+   
+   return eina_hash_del(desktop->x, key, NULL);
+}
