@@ -1,17 +1,50 @@
 /* vim: set sw=4 ts=4 sts=4 et: */
+
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+
+#include <stdio.h>
+#include <string.h>
+#include <ctype.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <sys/mman.h>
+
+#ifdef HAVE_ALLOCA_H
+# include <alloca.h>
+#elif defined __GNUC__
+# define alloca __builtin_alloca
+#elif defined _AIX
+# define alloca __alloca
+#elif defined _MSC_VER
+# include <malloc.h>
+# define alloca _alloca
+#else
+# include <stddef.h>
+# ifdef  __cplusplus
+extern "C"
+# endif
+void *alloca (size_t);
+#endif
+
 #include "Efreet.h"
 #include "efreet_private.h"
 
-static Ecore_Hash *efreet_ini_parse(const char *file);
+#ifdef EFREET_MODULE_LOG_DOM
+#undef EFREET_MODULE_LOG_DOM
+#endif
+#define EFREET_MODULE_LOG_DOM _efreet_ini_log_dom
+static int _efreet_ini_log_dom = -1;
+
+static Eina_Hash *efreet_ini_parse(const char *file);
 static char *efreet_ini_unescape(const char *str);
-
-static void efreet_ini_section_save(Ecore_Hash_Node *node, FILE *f);
-static void efreet_ini_value_save(Ecore_Hash_Node *node, FILE *f);
-
-/**
- * The number of times the Ini subsytem has been initialized
- */
-static int init = 0;
+static int _efreet_ini_log_om = -1;
+static Eina_Bool
+efreet_ini_section_save(const Eina_Hash *hash, const void *key, void *data, void *fdata);
+static Eina_Bool
+efreet_ini_value_save(const Eina_Hash *hash, const void *key, void *data, void *fdata);
 
 /**
  * @internal
@@ -21,9 +54,13 @@ static int init = 0;
 int
 efreet_ini_init(void)
 {
-    if (init++) return init;
-    if (!ecore_string_init()) return --init;
-    return init;
+    _efreet_ini_log_dom = eina_log_domain_register("Efreet_init", EFREET_DEFAULT_LOG_COLOR);
+    if (_efreet_ini_log_dom < 0)
+    {
+	ERROR("Efreet: Could not create a log domain for efreet_init");
+	return 0;
+    }
+    return 1;
 }
 
 /**
@@ -31,12 +68,10 @@ efreet_ini_init(void)
  * @returns the number of initializations left for this system
  * @brief Attempts to shut down the subsystem if nothing else is using it
  */
-int
+void
 efreet_ini_shutdown(void)
 {
-    if (--init) return init;
-    ecore_string_shutdown();
-    return init;
+    eina_log_domain_unregister(_efreet_ini_log_dom);
 }
 
 /**
@@ -54,6 +89,9 @@ efreet_ini_new(const char *file)
     ini = NEW(Efreet_Ini, 1);
     if (!ini) return NULL;
 
+    /* This can validly be NULL at the moment as _parse() will return NULL
+     * if the input file doesn't exist. Should we change _parse() to create
+     * the hash and only return NULL on failed parse? */
     ini->data = efreet_ini_parse(file);
 
     return ini;
@@ -62,179 +100,182 @@ efreet_ini_new(const char *file)
 /**
  * @internal
  * @param file The file to parse
- * @return Returns an Ecore_Hash with the contents of @a file, or NULL on failure
- * @brief Parses the ini file @a file into an Ecore_Hash
+ * @return Returns an Eina_Hash with the contents of @a file, or NULL if the
+ *         file fails to parse or if the file doesn't exist
+ * @brief Parses the ini file @a file into an Eina_Hash
  */
-static Ecore_Hash *
+static Eina_Hash *
 efreet_ini_parse(const char *file)
 {
+    const char *buffer, *line_start;
     FILE *f;
+    Eina_Hash *data, *section = NULL;
+    struct stat file_stat;
+    int line_length, left;
 
-    /* a static buffer for quick reading of lines that fit */
-    char static_buf[4096];
-    int static_buf_len = 4096;
-
-    /* a big buffer to allocate for lines that are larger than the static one */
-    char *big_buf = NULL;
-    int big_buf_len = 0;
-    int big_buf_step = static_buf_len;
-
-    /* the current location to read into (with fgets) and the amount to read */
-    char *read_buf;
-    int read_len;
-
-    /* the current buffer to parse */
-    char *buf;
-
-    Ecore_Hash *data, *section = NULL;
-
-    /* start with the static buffer */
-    buf = read_buf = static_buf;
-    read_len = static_buf_len;
+    if (!file) return NULL;
 
     f = fopen(file, "rb");
     if (!f) return NULL;
 
-    data = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-    ecore_hash_free_key_cb_set(data, ECORE_FREE_CB(ecore_string_release));
-    ecore_hash_free_value_cb_set(data, ECORE_FREE_CB(ecore_hash_destroy));
-
-    /* if a line is longer than the buffer size, this \n will get overwritten. */
-    read_buf[read_len - 2] = '\n';
-    while (fgets(read_buf, read_len, f) != NULL)
+    if (fstat(fileno(f), &file_stat) || (file_stat.st_size < 1))
     {
-        char *key, *value, *p;
-        char *sep;
+        fclose(f);
+        return NULL;
+    }
+    if (!S_ISREG(file_stat.st_mode)) /* if not a regular file - close */
+    {
+        fclose(f);
+        return NULL;
+    }
 
-        /* handle lines longer than the buffer size */
-        if (read_buf[read_len-2] != '\n')
+    left = file_stat.st_size;
+    buffer = mmap(NULL, left, PROT_READ, MAP_SHARED, fileno(f), 0);
+    if (buffer == MAP_FAILED)
+    {
+        fclose(f);
+        return NULL;
+    }
+
+    data = eina_hash_string_small_new(EINA_FREE_CB(eina_hash_free));
+
+    line_start = buffer;
+    while (left > 0)
+    {
+        int sep;
+
+        /* find the end of line */
+        for (line_length = 0;
+                (line_length < left) &&
+                (line_start[line_length] != '\n'); line_length++)
+            ;
+
+        /* check for all white space */
+        while (isspace(line_start[0]) && (line_length > 0))
         {
-            int len;
-            len = strlen(buf);
-
-            if (!big_buf)
-            {
-              /* create new big buffer and copy in contents of static buf */
-              big_buf_len = 2 * big_buf_step;
-              big_buf = malloc(big_buf_len * sizeof(char));
-              strncpy(big_buf, buf, len + 1);
-            }
-            else if (buf == big_buf)
-            {
-              /* already using the big buffer. increase its size for the next read */
-              big_buf_len += big_buf_step;
-              big_buf = realloc(big_buf, big_buf_len);
-            }
-            else
-            {
-              /* the big buffer exists, but we aren't using it yet. copy contents of static buf in */
-              strncpy(big_buf, buf, len);
-            }
-
-            /* use big_buffer for next fgets and subsequent parsing */
-            buf = big_buf;
-            read_buf = big_buf + len;
-            read_len = big_buf_len - len;
-            read_buf[read_len-2] = '\n';
-
-            continue;
+            line_start++;
+            line_length--;
         }
 
         /* skip empty lines and comments */
-        if (buf[0] == '\0' || buf[0] == '\n' || buf[0] == '#') goto next_line;
+        if ((line_length == 0) || (line_start[0] == '\r') ||
+                (line_start[0] == '\n') || (line_start[0] == '#') ||
+                (line_start[0] == '\0'))
+            goto next_line;
 
         /* new section */
-        if (buf[0] == '[')
+        if (line_start[0] == '[')
         {
-            char *header, *p;
-            header = buf + 1;
+            int header_length;
 
-            p = strchr(header, ']');
-            if (p)
+            /* find the ']' */
+            for (header_length = 1;
+                    (header_length < line_length) &&
+                    (line_start[header_length] != ']'); ++header_length)
+                ;
+
+            if (line_start[header_length] == ']')
             {
-                Ecore_Hash *old;
-                *p = '\0';
-                section = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-                ecore_hash_free_key_cb_set(section, ECORE_FREE_CB(ecore_string_release));
-                ecore_hash_free_value_cb_set(section, ECORE_FREE_CB(free));
+                const char *header;
 
-                old = ecore_hash_remove(data, header);
-                //if (old) printf("[efreet] Warning: duplicate section '%s' in file '%s'\n", header, file);
-                IF_FREE_HASH(old);
-                ecore_hash_set(data, (void *)ecore_string_instance(header),
-                                                                section);
+                header = alloca(header_length * sizeof(unsigned char));
+                if (!header) goto next_line;
+
+                memcpy((char*)header, line_start + 1, header_length - 1);
+                ((char*)header)[header_length - 1] = '\0';
+
+                section = eina_hash_string_small_new(free);
+
+                eina_hash_del(data, header, NULL);
+//                if (old) INF("[efreet] Warning: duplicate section '%s' "
+  //                              "in file '%s'", header, file);
+
+                eina_hash_add(data, header, section);
             }
             else
             {
                 /* invalid file - skip line? or refuse to parse file? */
                 /* just printf for now till we figure out what to do */
-                printf("Invalid file (%s) (missing ] on group name)\n", file);
+//                printf("Invalid file (%s) (missing ] on group name)\n", file);
             }
             goto next_line;
         }
 
-        /* parse key=value pair */
-        sep = strchr(buf, '=');
-        key = buf;
-
-        if (sep)
+        if (section == NULL)
         {
-            /* trim whitespace from end of key */
-            p = sep;
-            while (p > key && isspace(*(p - 1))) p--;
-            *p = '\0';
+//            INF("Invalid file (%s) (missing section)", file);
+            goto next_line;
+        }
 
-            value = sep + 1;
+        /* find for '=' */
+        for (sep = 0; (sep < line_length) && (line_start[sep] != '='); ++sep)
+            ;
+
+        if (sep < line_length)
+        {
+            const char *key, *value;
+            int key_end, value_start, value_end;
+
+            /* trim whitespace from end of key */
+            for (key_end = sep - 1;
+                    (key_end > 0) && isspace(line_start[key_end]); --key_end)
+                ;
+
+            if (!isspace(line_start[key_end])) key_end++;
 
             /* trim whitespace from start of value */
-            while (*value && isspace(*value)) value++;
+            for (value_start = sep + 1;
+                 (value_start < line_length) &&
+                 isspace(line_start[value_start]); ++value_start)
+                ;
 
             /* trim \n off of end of value */
-            p = value + strlen(value) - 1;
-            while (p > value && (*p == '\n' || *p == '\r')) p--;
-            *(p + 1) = '\0';
+            for (value_end = line_length;
+                 (value_end > value_start) &&
+                 ((line_start[value_end] == '\n') ||
+                  (line_start[value_end] == '\r')); --value_end)
+                ;
 
-            if (key && value && *key && *value)
+            if (line_start[value_end] != '\n'
+                    && line_start[value_end] != '\r'
+                    && value_end < line_length)
+                value_end++;
+
+            /* make sure we have a key. blank values are allowed */
+            if (key_end == 0)
             {
-                char *old;
+                /* invalid file... */
+//                INF("Invalid file (%s) (invalid key=value pair)", file);
 
-                old = ecore_hash_remove(section, key);
-                //if (old) printf("[efreet] Warning: duplicate key '%s' in file '%s'\n", key, file);
-                IF_FREE(old);
-
-                ecore_hash_set(section, (void *)ecore_string_instance(key),
-                               efreet_ini_unescape(value));
+                goto next_line;
             }
-        }
-        else
-        {
-            /* check if line is all whitespace, if so, skip it */
-            int nonwhite = 0;
-            p = buf;
-            while (*p)
-            {
-                if (!isspace(*p))
-                {
-                    nonwhite = 1;
-                    break;
-                }
-                p++;
-            }
-            if (!nonwhite) goto next_line;
 
-            /* invalid file... */
-            printf("Invalid file (%s) (missing = from key=value pair)\n", file);
+            key = alloca((key_end + 1) * sizeof(unsigned char));
+            value = alloca((value_end - value_start + 1) * sizeof(unsigned char));
+            if (!key || !value) goto next_line;
+
+            memcpy((char*)key, line_start, key_end);
+            ((char*)key)[key_end] = '\0';
+
+            memcpy((char*)value, line_start + value_start,
+                    value_end - value_start);
+            ((char*)value)[value_end - value_start] = '\0';
+
+            eina_hash_del(section, key, NULL);
+            eina_hash_add(section, key, efreet_ini_unescape(value));
         }
+//        else
+//        {
+//            /* invalid file... */
+//            INF("Invalid file (%s) (missing = from key=value pair)", file);
+//        }
 
 next_line:
-        /* finished parsing a line. use static buffer for next line */
-        buf = read_buf = static_buf;
-        read_len = static_buf_len;
-        read_buf[read_len - 2] = '\n';
+        left -= line_length + 1;
+        line_start += line_length + 1;
     }
-
+    munmap((char*) buffer, file_stat.st_size);
     fclose(f);
-    if (big_buf) free(big_buf);
 
     return data;
 }
@@ -267,7 +308,7 @@ efreet_ini_save(Efreet_Ini *ini, const char *file)
 
     f = fopen(file, "wb");
     if (!f) return 0;
-    ecore_hash_for_each_node(ini->data, ECORE_FOR_EACH(efreet_ini_section_save), f);
+    eina_hash_foreach(ini->data, efreet_ini_section_save, f);
     fclose(f);
 
     return 1;
@@ -284,7 +325,7 @@ efreet_ini_section_set(Efreet_Ini *ini, const char *section)
 {
     if (!ini || !ini->data || !section) return 0;
 
-    ini->section = ecore_hash_get(ini->data, section);
+    ini->section = eina_hash_find(ini->data, section);
     return (ini->section ? 1 : 0);
 }
 
@@ -297,22 +338,16 @@ efreet_ini_section_set(Efreet_Ini *ini, const char *section)
 EAPI void
 efreet_ini_section_add(Efreet_Ini *ini, const char *section)
 {
-    Ecore_Hash *hash;
+    Eina_Hash *hash;
 
     if (!ini || !section) return;
 
     if (!ini->data)
-    {
-        ini->data = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-        ecore_hash_free_key_cb_set(ini->data, ECORE_FREE_CB(ecore_string_release));
-        ecore_hash_free_value_cb_set(ini->data, ECORE_FREE_CB(ecore_hash_destroy));
-    }
-    if (ecore_hash_get(ini->data, section)) return;
+      ini->data = eina_hash_string_small_new(EINA_FREE_CB(eina_hash_free));
+    if (eina_hash_find(ini->data, section)) return;
 
-    hash = ecore_hash_new(ecore_str_hash, ecore_str_compare);
-    ecore_hash_free_key_cb_set(hash, ECORE_FREE_CB(ecore_string_release));
-    ecore_hash_free_value_cb_set(hash, ECORE_FREE_CB(free));
-    ecore_hash_set(ini->data, (void *)ecore_string_instance(section), hash);
+    hash = eina_hash_string_small_new(free);
+    eina_hash_add(ini->data, section, hash);
 }
 
 /**
@@ -327,7 +362,7 @@ efreet_ini_string_get(Efreet_Ini *ini, const char *key)
 {
     if (!ini || !key || !ini->section) return NULL;
 
-    return ecore_hash_get(ini->section, key);
+    return eina_hash_find(ini->section, key);
 }
 
 /**
@@ -342,7 +377,8 @@ efreet_ini_string_set(Efreet_Ini *ini, const char *key, const char *value)
 {
     if (!ini || !key || !ini->section) return;
 
-    ecore_hash_set(ini->section, (void *)ecore_string_instance(key), strdup(value));
+    eina_hash_del(ini->section, key, NULL);
+    eina_hash_add(ini->section, key, strdup(value));
 }
 
 /**
@@ -493,29 +529,28 @@ efreet_ini_localestring_get(Efreet_Ini *ini, const char *key)
     {
         snprintf(buf, maxlen, "%s[%s_%s@%s]", key, lang, country, modifier);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found && lang && country)
     {
         snprintf(buf, maxlen, "%s[%s_%s]", key, lang, country);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found && lang && modifier)
     {
         snprintf(buf, maxlen, "%s[%s@%s]", key, lang, modifier);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found && lang)
     {
         snprintf(buf, maxlen, "%s[%s]", key, lang);
         val = efreet_ini_string_get(ini, buf);
-        if (val) found = 1;
-
+        if (val && (*val != '\0')) found = 1;
     }
 
     if (!found)
@@ -568,6 +603,20 @@ efreet_ini_localestring_set(Efreet_Ini *ini, const char *key, const char *value)
     FREE(buf);
 }
 
+/**
+ * @param ini: The ini struct to work with
+ * @param key: The key to remove
+ * @return Returns no value
+ * @brief Remove the given key from the ini struct
+ */
+EAPI void
+efreet_ini_key_unset(Efreet_Ini *ini, const char *key)
+{
+    if (!ini || !key || !ini->section) return;
+
+    eina_hash_del(ini->section, key, NULL);
+}
+
 /**
  * @param str The string to unescape
  * @return An allocated unescaped string
@@ -587,7 +636,7 @@ efreet_ini_unescape(const char *str)
     dest = buf;
     while(*p)
     {
-        if (*p == '\\')
+        if ((*p == '\\') && (p[1] != '\0'))
         {
             p++;
             switch (*p)
@@ -622,15 +671,21 @@ efreet_ini_unescape(const char *str)
     return buf;
 }
 
-static void
-efreet_ini_section_save(Ecore_Hash_Node *node, FILE *f)
+static Eina_Bool
+efreet_ini_section_save(const Eina_Hash *hash __UNUSED__, const void *key, void *value, void *fdata)
 {
-    fprintf(f, "[%s]\n", (char *)node->key);
-    ecore_hash_for_each_node(node->value, ECORE_FOR_EACH(efreet_ini_value_save), f);
+    FILE *f = fdata;
+
+    fprintf(f, "[%s]\n", (char *)key);
+    eina_hash_foreach(value, efreet_ini_value_save, f);
+    return EINA_TRUE;
 }
 
-static void
-efreet_ini_value_save(Ecore_Hash_Node *node, FILE *f)
+static Eina_Bool
+efreet_ini_value_save(const Eina_Hash *hash __UNUSED__, const void *key, void *value, void *fdata)
 {
-    fprintf(f, "%s=%s\n", (char *)node->key, (char *)node->value);
+    FILE *f = fdata;
+
+    fprintf(f, "%s=%s\n", (char *)key, (char *)value);
+    return EINA_TRUE;
 }
