@@ -42,11 +42,11 @@ struct Efreet_Xml
 
     Efreet_Xml_Attribute **attributes;  /**< The attributes for this node */
 
-    Ecore_DList *children;               /**< Child nodes */
+    Eina_List *children;               /**< Child nodes */
 };
 
 int efreet_xml_init(void);
-int efreet_xml_shutdown(void);
+void efreet_xml_shutdown(void);
 
 Efreet_Xml *efreet_xml_new(const char *file);
 void efreet_xml_del(Efreet_Xml *xml);
